unit SegmentedMemoryStream;

(*
  Unit SegmentedMemoryStream

  Copyright (C) 2008,2017 by Thomas Voelker (thomas.voelker@voelker-hud.de)

  Licensed under the terms of the GPL or LGPL.

*)

interface

uses Classes,SysUtils;

type

  TSegmentSize = ({ss64b = 6,}ss64kB = 16,ss128kB = 17,ss256kB=18,ss512kB=19,ss1MB=20,ss2MB=21);


  TSegMemStream = Class(TStream)
    private
      FSize:Int64;
      FMemory:array of Pointer;
      FSegmentSize:TSegmentSize;
      FPosition:Int64;
    protected
      function GetSize: Int64; override;                 //ok
      procedure SetSize(NewSize: Longint); overload; override; deprecated;//ok
      procedure SetSize(const NewSize: Int64); overload; override;//ok
//unfortunately the untyped buffer cannot be transitioned to TBytes without allocating memory for TBytes because
//TBytes itself is a pointer to the first byte and compiler handles preceding DWord values (number of elements @ offset -4,
//reference counter @ offset -8) which would have to be set up before the buffer - but the memory on those positions may
//already be aquired. That made me write the functions WriteBuffer64 and ReadBuffer64 that are internally called by the overrides. The reason
//for this was, that I did not want to write the same code twice.
      function ReadBuffer64(var Buffer;Offset,Count:Int64):Int64; //ok
      function WriteBuffer64(const Buffer;Offset,Count:Int64):Int64; //ok

    public
      Constructor Create; Overload; //ok
      Constructor Create(const InitialSize:Int64;SegmentSize:TSegmentSize;ClearBuffer:Boolean);overload; //ok
      destructor Destroy; override; //ok
      procedure Clear; //ok

      function LoadFromStream(Stream: TStream):Boolean;
      function SaveToStream(Stream: TStream):Boolean;
      procedure LoadFromFile(const FileName: string);
      procedure SaveToFile(const FileName: string);

      function Read(var Buffer; Count: Longint): Longint; overload; override; //ok
      function Read(Buffer: TBytes; Offset, Count: Longint): Longint; overload; override; //ok
      function Read64(Buffer: TBytes; Offset, Count: Int64): Int64; override; //ok

      function Write(const Buffer; Count: Longint): Longint; overload; override; //ok
      function Write(const Buffer: TBytes; Offset, Count: Longint): Longint; overload; override; //ok
      function Write64(const Buffer: TBytes; Offset, Count: Int64): Int64; override; //ok

      function Seek(Offset: Longint; Origin: Word): Longint; overload; override;           //ok
      function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; overload; override;      //ok
  End;





implementation
  uses Dialogs;

//TFileObjectCache
  Constructor TSegMemStream.Create;
  begin
    Create(0,ss1MB,False);
  end;

  Constructor TSegMemStream.Create(const InitialSize:Int64;SegmentSize:TSegmentSize;ClearBuffer:Boolean);
  begin
    inherited Create;
    FSegmentSize:=SegmentSize;
    SetSize(InitialSize);
    If ClearBuffer Then Clear;
  end;

  procedure TSegMemStream.SetSize(NewSize: Integer);
  var Size64:Int64;
  begin
    Size64:=NewSize;
    Setsize(Size64);
  end;

  procedure TSegMemStream.SetSize(const NewSize: Int64);
  var i,NewNumSegments,OldNumSegments:Cardinal;
      ShiftBits,DivBits:Byte;
      RemainingSize:Int64;
      OldLastSegmentBytes,NewLastSegmentBytes,SegmentBytes:Int64;
      pOld,pNew:Pointer;
  begin
    If NewSize=Size Then Exit;
    ShiftBits:=Ord(FSegmentSize);
    DivBits:=(SizeOf(NewSize) shl 3)-ShiftBits;
    NewNumSegments:=NewSize shr ShiftBits;
    NewLastSegmentBytes:=(NewSize shl DivBits)shr DivBits;

    if NewLastSegmentBytes>0 then Inc(NewNumSegments);
    OldNumSegments:=Length(FMemory);
    OldLastSegmentBytes:=(FSize shl DivBits)shr DivBits;
    SegmentBytes:=1 shl ShiftBits;
//    ShowMessage(Format('NewSize: %d, NewNumSegments: %d, SegmentBytes: %d, LastSegmentBytes: %d',[NewSize, NewNumSegments,SegmentBytes,NewLastSegmentBytes]));

    If OldNumSegments=NewNumSegments Then Begin
      pOld:=FMemory[OldNumSegments-1];
      GetMem(pNew,NewLastSegmentBytes);
      if(NewLastSegmentBytes<OldLastSegmentBytes)Then Move(pOld^,pNew^,NewLastSegmentBytes)
      Else Move(pOld^,pNew^,OldLastSegmentBytes);
      FMemory[OldNumSegments-1]:=pNew;
      FreeMem(pOld,OldLastSegmentBytes);
    End else if OldNumSegments<NewNumSegments then Begin
      SetLength(FMemory,NewNumSegments);
      if OldLastSegmentBytes>0 Then Begin
        pOld:=FMemory[OldNumSegments-1];
        GetMem(pNew,SegmentBytes);
        Move(pOld^,pNew^,OldLastSegmentBytes);
        FMemory[OldNumSegments-1]:=pNew;
        FreeMem(pOld,OldLastSegmentBytes);
      End;
      if NewLastSegmentBytes=0 Then For i:=OldNumSegments To NewNumSegments-1 Do Begin
        GetMem(pNew,SegmentBytes);
        FMemory[i]:=pNew;
      End else Begin
        if NewNumSegments>(OldNumSegments+1)Then For i:=OldNumSegments To NewNumSegments-2 Do Begin
          GetMem(pNew,SegmentBytes);
          FMemory[i]:=pNew;
        End;
        GetMem(pNew,NewLastSegmentBytes);
        FMemory[NewNumSegments-1]:=pNew;
      End;
    End else Begin
      For i:=(OldNumSegments - 1) DownTo (NewNumSegments)Do Begin
        pOld:=FMemory[i];
        if (i=OldNumSegments-1)And(OldLastSegmentBytes>0)Then FreeMem(pOld,OldLastSegmentBytes)
        Else FreeMem(pOld,SegmentBytes);
        FMemory[i]:=nil;
      End;
      SetLength(FMemory,NewNumSegments);
      if (NewNumSegments>0)And(NewLastSegmentBytes<>SegmentBytes) Then Begin
        pOld:=FMemory[NewNumSegments-1];
        GetMem(pNew,NewLastSegmentBytes);
        Move(pOld^,pNew^,NewLastSegmentBytes);
        FMemory[i]:=pNew;
        FreeMem(pOld,SegmentBytes);
      End;
    End;
    FSize:=NewSize;
  end;


procedure TSegMemStream.Clear;
const cFillVal:Cardinal = 0;
      cSecSizeDiv=SizeOf(cFillVal);
      cFillByte:Byte=0;
var i:Cardinal;
    j:Int64;
    LastSegmentBytes:int64;
    ShiftBits,DivBits:Byte;
    Chunks,SegmentAddr:Int64;
    p:Pointer;
    SegmentCount:Cardinal;

begin
  ShiftBits:=Ord(FSegmentSize);
  DivBits:=(SizeOf(FSize) shl 3)-ShiftBits;
  LastSegmentBytes:=(FSize shl DivBits)shr DivBits;
  Chunks:=(1 shl ShiftBits) div cSecSizeDiv;
  SegmentCount:=Length(FMemory);
  If LastSegmentBytes=0Then Begin
    For i:=0 to SegmentCount-1Do Begin
      SegmentAddr:=Int64(FMemory[i]);
      For j:=0To Chunks-1Do Begin
        p:=Pointer(SegmentAddr+cSecSizeDiv * j);
        Move(cFillVal,p^,cSecSizeDiv);
//        Move(i,p^,cSecSizeDiv);
      End;
    End;
  End Else Begin
    if SegmentCount>1 Then For i:=0To SegmentCount-2 Do Begin
      SegmentAddr:=Int64(FMemory[i]);
      For j:=0To Chunks-1Do Begin
        p:=Pointer(SegmentAddr+cSecSizeDiv * j);
        Move(cFillVal,p^,cSecSizeDiv);
//        Move(i,p^,cSecSizeDiv);
      End;
    End;
    SegmentAddr:=Int64(FMemory[SegmentCount-1]);
    Chunks:=LastSegmentBytes div cSecSizeDiv;
    For j:=0To Chunks-1 Do Begin
      p:=Pointer(SegmentAddr + cSecSizeDiv * j);
      Move(cFillVal,p^,cSecSizeDiv);
//      Move(i,p^,cSecSizeDiv);
    End;
    If LastSegmentBytes mod cSecSizeDiv >0 Then Begin
      For j:=0To (LastSegmentBytes mod cSecSizeDiv)-1 Do Begin
        p:=Pointer(SegmentAddr + cSecSizeDiv * Chunks + j);
        Move(cFillByte,p^,1);
      End;
    End;
  End;
end;

Destructor TSegMemStream.Destroy;
begin
  SetSize(0);
  inherited Destroy;
end;

Function TSegMemStream.GetSize:Int64;
begin
  Result:=FSize;
end;



function TSegMemStream.Seek( Offset: LongInt; Origin: Word): LongInt;
var Offset64,Result64:Int64;
    Origin64:TSeekOrigin;
begin
  Offset64:=Offset;
  Case Origin of
    soFromBeginning:Origin64:=soBeginning;
    soFromCurrent:Origin64:=soCurrent;
    soFromEnd:Origin64:=soEnd;
  End;
  Result64:=Seek(Offset64,Origin64);
  Result:=Result64;
end;

function TSegMemStream.Seek(const Offset: Int64; Origin: TSeekOrigin):Int64;
begin
  Case Origin of
    soBeginning:FPosition := Offset;
    soCurrent:FPosition:= FPosition + Offset;
    soEnd:FPosition := FSize + Offset;
  End;
  Result:=FPosition;
end;

function TSegMemStream.ReadBuffer64(var Buffer; Offset: Int64; Count: Int64):Int64;
var StartSegment,EndSegment:Cardinal;
    StartPosition,EndPosition:Int64;
    BytesRead:Int64;
    ShiftBits,DivBits:Byte;
    SegmentPointer,BufferPointer:Pointer;
    SegmentAddr,BufferAddr:Int64;
    i:Int64;
    SegmentSize:Int64;
begin
  If Count>FSize-FPosition+1 Then Count:=FSize-FPosition+1;
  Result:=Count;
  If Count=0 Then Exit;

  ShiftBits:=Ord(FSegmentSize);
  DivBits:=(SizeOf(FPosition) shl 3) - ShiftBits;
  StartSegment:=FPosition shr ShiftBits;
  StartPosition:=(FPosition shl DivBits)shr DivBits;

  EndSegment:=(FPosition+Count-1)shr ShiftBits;
  EndPosition:=((FPosition+Count-1) shl DivBits)shr DivBits;

  BufferAddr:=Int64(Addr(Buffer))+Offset;
  BufferPointer:=Pointer(BufferAddr);

  SegmentAddr:=Int64(FMemory[StartSegment]);
  SegmentPointer:=Pointer(SegmentAddr + StartPosition);
  SegmentSize:=1 shl ShiftBits;
  If StartSegment=EndSegment Then Begin
    Move(SegmentPointer^,BufferPointer^,Count);
    Inc(FPosition,Count);
    Exit;
  End Else Begin
    Move(SegmentPointer^,BufferPointer^,SegmentSize-StartPosition);
    BytesRead:=SegmentSize-StartPosition;
  End;
  If StartSegment+1<EndSegment Then Begin
    For i:=StartSegment+1 To EndSegment-1 Do Begin
      SegmentPointer:=FMemory[i];
      BufferPointer:=Pointer(BufferAddr+BytesRead);
      Move(SegmentPointer^,BufferPointer^,SegmentSize);
      Inc(BytesRead,SegmentSize);
    End;
  End;
  SegmentPointer:=FMemory[EndSegment];
  BufferPointer:=Pointer(BufferAddr+BytesRead);
  Move(SegmentPointer^,BufferPointer^,EndPosition+1);
  Inc(FPosition,Count);
end;

function TSegMemStream.Read64(Buffer: TBytes; Offset: Int64; Count: Int64):Int64;
begin
  If Length(Buffer)<Count Then Count:=Length(Buffer);
  Result:=ReadBuffer64(Buffer[0],Offset,Count);
end;

function TSegMemStream.WriteBuffer64(const Buffer; Offset: Int64; Count: Int64):Int64;
var StartSegment,EndSegment:Cardinal;
    StartPosition,EndPosition:Int64;
    BytesWritten:Int64;
    ShiftBits,DivBits:Byte;
    SegmentPointer,BufferPointer:Pointer;
    SegmentAddr,BufferAddr:Int64;
    i:Int64;
    SegmentSize:Int64;
begin
  if FPosition+Count-1>FSize Then SetSize(FPosition+Count-1);
  Result:=Count;

  ShiftBits:=Ord(FSegmentSize);
  DivBits:=(SizeOf(FPosition) shl 3) - ShiftBits;
  StartSegment:=FPosition shr ShiftBits;
  StartPosition:=(Fposition shl DivBits)shr DivBits;

  EndSegment:=(FPosition+Count-1)shr ShiftBits;
  EndPosition:=((FPosition+Count-1) shl DivBits)shr DivBits;

  BufferAddr:=Int64(Addr(Buffer))+Offset;
  BufferPointer:=Pointer(BufferAddr);

  SegmentAddr:=Int64(FMemory[StartSegment]);
  SegmentPointer:=Pointer(SegmentAddr + StartPosition);
  SegmentSize:=1 shl ShiftBits;
  If StartSegment=EndSegment Then Begin
    Move(BufferPointer^,SegmentPointer^,Count);
    Inc(FPosition,Count);
    Exit;
  End Else Begin
    Move(BufferPointer^,SegmentPointer^,SegmentSize-StartPosition);
    BytesWritten:=SegmentSize-StartPosition;
  End;
  If StartSegment+1<EndSegment Then Begin
    For i:=StartSegment+1 To EndSegment-1 Do Begin
      SegmentPointer:=FMemory[i];
      BufferPointer:=Pointer(BufferAddr+BytesWritten);
      Move(BufferPointer^,SegmentPointer^,SegmentSize);
      Inc(BytesWritten,SegmentSize);
    End;
  End;
  SegmentPointer:=FMemory[EndSegment];
  BufferPointer:=Pointer(BufferAddr+BytesWritten);
  Move(BufferPointer^,SegmentPointer^,EndPosition+1);
  Inc(FPosition,Count);
end;

function TSegMemStream.Write64(const Buffer: TBytes; Offset: Int64; Count: Int64):Int64;
begin
  if Count>Length(Buffer)Then Count:=Length(Buffer);
  Result:=WriteBuffer64(Buffer[0],Offset,Count);
end;

function TSegMemStream.Read(var Buffer; Count: Integer):LongInt;
begin
  Result:=ReadBuffer64(Buffer,0,Count);
end;

function TSegMemStream.Read(Buffer: TBytes; Offset: Integer; Count: Integer):LongInt;
begin
  Result:=Read64(Buffer,Offset,Count);
end;

function TsegMemStream.Write(const Buffer; Count: Integer):LongInt;
begin
  Result:=WriteBuffer64(Buffer,0,Count);
end;

function TSegMemStream.Write(const Buffer: TBytes; Offset: Integer; Count: Integer):LongInt;
begin
  Result:=Write64(Buffer,Offset,Count);
end;


function TSegMemStream.LoadFromStream(Stream: TStream):Boolean;
begin
  Result:=False;
  If Not Assigned(Stream) Then Exit;
  SetSize(Stream.Size);
  Stream.Position:=0;
  Position:=0;
  CopyFrom(Stream,FSize);
  Result:=True;
end;

function TSegMemStream.SaveToStream(Stream: TStream):Boolean;
begin
  Result:=False;
  If Not Assigned(Stream) Then Exit;
  Stream.Size:=Size;
  Stream.Position:=0;
  Position:=0;
  Stream.CopyFrom(Self,FSize);
  Result:=True;
end;

procedure TSegMemStream.LoadFromFile(const FileName: string);
var FS:TFileStream;
begin
  FS:=TFileStream.Create(FileName,fmOpenRead+fmShareDenyWrite);
  try
    LoadFromStream(FS);
  finally
    FreeAndNil(FS);
  end;
end;

procedure TSegMemStream.SaveToFile(const FileName: string);
var FS:TFileStream;
begin
  FS:=TFileStream.Create(FileName,fmCreate+fmShareDenyWrite);
  try
    SaveToStream(FS);
  finally
    FreeAndNil(FS);
  end;
end;


end.
